local class = require "engine.class"
local ActorTalents = require "engine.interface.ActorTalents"
local Object = require "mod.class.Object"

class:bindHook("ToME:load", function (self, data)
	ActorTalents:loadDefinition("/data-esp-mod/talents/gifts/call.lua")
	ActorTalents:loadDefinition("/data-esp-mod/talents/psionic/mentalism.lua")
end)

-- Modify ESP fields for loaded objects
local function updateObjectEsp(objects, object_esps)
	for object_name, object_esp in pairs(object_esps) do
		for _, object in ipairs(objects) do
			if object.name == object_name then
				for _, field in ipairs({"wielder", "imbue_powers"}) do
					if object[field] and (object[field].esp_all or object[field].esp_range or object[field].esp) then
						object[field].esp_all, object[field].esp_range = nil, nil
						object[field].esp = object_esp
					end
				end

				break
			end
		end
	end
end

-- Replacement for artifact use() function
local function stone_guardian(self, who)
	local x, y = util.findFreeGrid(who.x, who.y, 5, true, {[engine.Map.ACTOR]=true})
	if not x then
		game.logPlayer(self, "Not enough space to invoke!")
		return
	end
	local Talents = require "engine.interface.ActorTalents"
	local NPC = require "mod.class.NPC"
	local m = NPC.new{
		resolvers.nice_tile{image="invis.png", add_mos = {{image="terrain/darkgreen_moonstone_01.png", display_h=2, display_y=-1}}},
		name = "Stone Guardian",
		type = "totem", subtype = "antimagic",
		desc = "This massive stone pillar drips with a viscious slime. Nature's power flows through it, obliterating magic all around it..",
		rank = 3,
		blood_color = colors.GREEN,
		display = "T", color=colors.GREEN,
		life_rating=18,
		combat = {
			dam=resolvers.rngavg(50,60),
			atk=resolvers.rngavg(50,75), apr=25,
			dammod={wil=1.1}, physcrit = 10,
			damtype=engine.DamageType.SLIME,
		},
		level_range = {1, nil}, exp_worth = 0,
		silent_levelup = true,
		combat_armor=50,
		combat_armor_hardiness=70,
		autolevel = "wildcaster",
		ai = "summoned", ai_real = "dumb_talented_simple", ai_state = { talent_in=1, },
		never_move=true,
		stats = { str=14, dex=18, mag=10, con=12, wil=20, cun=20, },
		size_category = 5,
		blind=1,
		esp = {"all"},
		resists={[engine.DamageType.BLIGHT] = 40, [engine.DamageType.ARCANE] = 40, [engine.DamageType.NATURE] = 70},
		no_breath = 1,
		cant_be_moved = 1,
		stone_immune = 1,
		confusion_immune = 1,
		fear_immune = 1,
		teleport_immune = 1,
		disease_immune = 1,
		poison_immune = 1,
		stun_immune = 1,
		blind_immune = 1,
		cut_immune = 1,
		knockback_resist,
		combat_mentalresist=50,
		combat_spellresist=100,
		on_act = function(self) self:project({type="ball", range=0, radius=5, selffire=false}, self.x, self.y, DamageType.SILENCE, {dur=2, power_check=self:combatMindpower()}) end,
		resolvers.talents{
			[Talents.T_RESOLVE]={base=3, every=6},
			[Talents.T_MANA_CLASH]={base=3, every=5},
			[Talents.T_STUN]={base=3, every=4},
			[Talents.T_OOZE_SPIT]={base=5, every=4},
			[Talents.T_TENTACLE_GRAB]={base=1, every=6,},
		},
		
		faction = who.faction,
		summoner = who, summoner_gain_exp=true,
		summon_time=15,
	}

	m:resolve()
	game.zone:addEntity(game.level, m, "actor", x, y)
	m.remove_from_party_on_death = true,
	game.party:addMember(m, {
		control=false,
		type="summon",
		title="Summon",
		orders = {target=true, leash=true, anchor=true, talents=true},
	})
	return {id=true, used=true}
end

class:bindHook("Entity:loadList", function (self, data)
	local file = (data.file:find('/')) == 1 and data.file or '/'..data.file
	-- Random artifacts require {type = range} format even for default ranges
	local random_fixups = {
		["telepathy"] = {all = 4.15},
		["orc telepathy"] = {["humanoid/orc"] = 8.35},
		["dragon telepathy"] = {dragon = 10},
		["demon telepathy"] = {["demon/minor"] = 10, ["demon/major"] = 10},
	}
	local item_fixups = {
		["/data/general/objects/boss-artifacts.lua"] = {
			["Dragonskull Helm"] = "dragon",
			["Chromatic Harness"] = "dragon",
			["Whip of Urh'Rok"] = {"demon/minor", "demon/major"},
		},
		["/data/general/objects/random-artifacts.lua"]         = random_fixups,
		["/data/general/objects/random-artifacts/generic.lua"] = random_fixups,
		["/data/general/objects/random-artifacts/ranged.lua"]  = random_fixups,
		["/data/general/objects/world-artifacts.lua"] = {
			["Genocide"] = "humanoid/orc",
			["Orc Feller"] = "humanoid/orc",
			["Unflinching Eye"] = {horror = 20},
			["The Inner Eye"] = {all = 7},
		},
		["/data/general/objects/world-artifacts-far-east.lua"] = {
			["Goedalath Rock"] = {"demon/major", "demon/minor"},
			["Blood-Edge"] = "undead/blood",
		},
	}
	local random_all = { theme={psionic=true}, name="telepathy", points = 60, rarity = 65, level_range = {1, 50}, wielder = {esp = {all = 4.15}}}
	if item_fixups[file] then updateObjectEsp(data.res, item_fixups[file]) end

	if file == "/data/general/objects/world-artifacts.lua" then
		for _, object in ipairs(data.res) do
			if object.name == "The Guardian's Totem" then
				-- gotta replace a whole function for the sake of one line
				object.use_power.use = stone_guardian
			end
		end
		-- Add "Brass Knuckle" to the list
		Object:loadList("/data-esp-mod/general/objects/new-artifact.lua", data.no_default, data.res, data.mod, data.loaded)
	elseif file == "/data/general/objects/random-artifacts.lua" then	-- add All type back in
		Object:loadList("/data-esp-mod/general/objects/random-artifacts.lua", data.no_default, data.res, data.mod, data.loaded)
	end
end)