local _M = loadPrevious(...)

local Esp = require "mod.class.interface.ActorEsp"

-- Import ESP interface functions into Actor namespace
_M._BASES[#_M._BASES] = Esp
local ignore_names = {init = true, _NAME = true, _PACKAGE = true, new = true}
for name, value in pairs(Esp) do
	if not ignore_names[name] then
		_M[name] = value
	end
end


-- Bypass old ESP initialization and do the new
local oldinit = _M.init
function _M:init(t, no_default)
	local oldesp, oldrange = t.esp, t.esp_range
	oldinit(self, t, no_default)
	t.esp, t.esp_range = oldesp, oldrange
	Esp.init(self, t)
end

-- Add item ESP effects cleanly
function _M:onWear(o, bypass_set)
	o.wielded = {}

	if self.player then o:forAllStack(function(so) so.__transmo = false end) end

	if o.set_list and not bypass_set then
		local list = {}
		for i, d in ipairs(o.set_list) do
			local po, item, inven_id = self:findInAllInventoriesBy(d[1], d[2])
			if po and self:getInven(inven_id).worn then
				list[#list+1] = po
			end
		end
		if #list == #o.set_list then
			for i, po in ipairs(list) do
				self:onTakeoff(po, true)
				po:check("on_set_complete", self)
				self:onWear(po, true)
				po.set_complete = true
			end
			o:check("on_set_complete", self)
			o.set_complete = true
		end
	end

	o:check("on_wear", self)
	-- Cut "learn_talent" and "esp" out separately so self (the Actor) isn't
	-- polluted with useless temporary value cruft.
	if o.wielder then
		for k, e in pairs(o.wielder) do
			if k == "learn_talent" then
				for tid, level in pairs(o.wielder.learn_talent) do
					self:learnItemTalent(o, tid, level)
				end
			elseif k == "esp" then
				self:addEsp(e)
			else
				o.wielded[k] = self:addTemporaryValue(k, e)
			end
		end
	end

	if o.talent_on_spell then
		self.talent_on_spell = self.talent_on_spell or {}
		for i = 1, #o.talent_on_spell do
			local id = util.uuid()
			self.talent_on_spell[id] = o.talent_on_spell[i]
			o.talent_on_spell[i]._id = id
		end
	end

	if o.talent_on_wild_gift then
		self.talent_on_wild_gift = self.talent_on_wild_gift or {}
		for i = 1, #o.talent_on_wild_gift do
			local id = util.uuid()
			self.talent_on_wild_gift[id] = o.talent_on_wild_gift[i]
			o.talent_on_wild_gift[i]._id = id
		end
	end

	if o.talent_on_mind then
		self.talent_on_mind = self.talent_on_mind or {}
		for i = 1, #o.talent_on_mind do
			local id = util.uuid()
			self.talent_on_mind[id] = o.talent_on_mind[i]
			o.talent_on_mind[i]._id = id
		end
	end

	-- Apply any special cursed logic
	if self:knowTalent(self.T_DEFILING_TOUCH) then
		local t = self:getTalentFromId(self.T_DEFILING_TOUCH)
		t.on_onWear(self, t, o)
	end

	-- Apply antimagic
	if o.power_source and o.power_source.antimagic then
		self:attr("spellpower_reduction", 1)
		self:attr("spell_failure", (o.material_level or 1) * 10)
	end

	-- Apply Psychometry
	if self:knowTalent(self.T_PSYCHOMETRY) then
		local t = self:getTalentFromId(self.T_PSYCHOMETRY)
		t.updatePsychometryCount(self, t)
	end

	self:breakReloading()

	self:checkMindstar(o)

	self:updateModdableTile()
	if self == game.player then game:playSound("actions/wear") end
end

-- Remove item ESP effects
function _M:onTakeoff(o, bypass_set)
	engine.interface.ActorInventory.onTakeoff(self, o)

	if o.set_list and o.set_complete and not bypass_set then
		local list = {}
		for i, d in ipairs(o.set_list) do
			local po, item, inven_id = self:findInAllInventoriesBy(d[1], d[2])
			if po then
				self:onTakeoff(po, true)
				po:check("on_set_broken", self)
				if po._special_set then
					for k, id in pairs(po._special_set) do
						po:removeTemporaryValue(k, id)
					end
					po._special_set = nil
				end
				self:onWear(po, true)
				po.set_complete = nil
			end
		end
		o:check("on_set_broken", self)
		if o._special_set then
			for k, id in pairs(o._special_set) do o:removeTemporaryValue(k, id) end
			o._special_set = nil
		end
		o.set_complete = nil
	end

	if o._special_wear then
		for k, id in pairs(o._special_wear) do o:removeTemporaryValue(k, id) end
		o._special_wear = nil
	end

	if o.talent_on_spell then
		self.talent_on_spell = self.talent_on_spell or {}
		for i = 1, #o.talent_on_spell do
			local id = o.talent_on_spell[i]._id
			self.talent_on_spell[id] = nil
		end
	end

	if o.talent_on_wild_gift then
		self.talent_on_wild_gift = self.talent_on_wild_gift or {}
		for i = 1, #o.talent_on_wild_gift do
			local id = o.talent_on_wild_gift[i]._id
			self.talent_on_wild_gift[id] = nil
		end
	end


	if o.talent_on_mind then
		self.talent_on_mind = self.talent_on_mind or {}
		for i = 1, #o.talent_on_mind do
			local id = o.talent_on_mind[i]._id
			self.talent_on_mind[id] = nil
		end
	end

	-- apply any special cursed logic
	if self:knowTalent(self.T_DEFILING_TOUCH) then
		local t = self:getTalentFromId(self.T_DEFILING_TOUCH)
		t.on_onTakeOff(self, t, o)
	end

	-- Apply antimagic
	if o.power_source and o.power_source.antimagic then
		self:attr("spellpower_reduction", -1)
		self:attr("spell_failure", -(o.material_level or 1) * 10)
	end

	-- Apply Psychometry
	if self:knowTalent(self.T_PSYCHOMETRY) then
		local t = self:getTalentFromId(self.T_PSYCHOMETRY)
		t.updatePsychometryCount(self, t)
	end

	-- Remove item talents and ESP effects
	if o.wielder then
		if o.wielder.learn_talent then
			for tid, level in pairs(o.wielder.learn_talent) do
				self:unlearnItemTalent(o, tid, level)
			end
		end
		if o.wielder.esp then
			self:removeEsp(o.wielder.esp)
		end
	end

	self:checkMindstar(o)

	self:breakReloading()

	self:updateModdableTile()
	if self == game.player then game:playSound("actions/takeoff") end
end

-- Make sure the new ESP can sense in the module
function _M:canSeeNoCache(actor, def, def_pct)
	if not actor then return false, 0 end

	-- Check for new ESP.
	if self:canSense(actor) then
		return true, 100
	end

	-- Blindness means can't see anything
	if self:attr("blind") then
		return false, 0
	end

	-- Check for stealth. Checks against the target cunning and level
	if actor:attr("stealth") and actor ~= self then
		local def = self:combatSeeStealth()
		local hit, chance = self:checkHitOld(def, actor:attr("stealth") + (actor:attr("inc_stealth") or 0), 0, 100)
		if not hit then
			return false, chance
		end
	end

	-- Check for invisibility. This is a "simple" checkHit between invisible and see_invisible attrs
	if actor:attr("invisible") then
		-- Special case, 0 see invisible, can NEVER see invisible things
		local def = self:combatSeeInvisible()
		if def <= 0 then return false, 0 end
		local hit, chance = self:checkHitOld(def, actor:attr("invisible"), 0, 100)
		if not hit then
			return false, chance
		end
	end

	if def ~= nil then
		return def, def_pct
	else
		return true, 100
	end
end

return _M